#include <stdio.h>
#include <tchar.h>
#include <stdarg.h>
#include <vector>
#include <windows.h>
#include "resource.h"

//--------------------------------------------------------------------------
static HINSTANCE g_hinst;

//--------------------------------------------------------------------------
static const unsigned char x1[] = 
{
  0x24,0x10,0x00,0x00,0x00, // +00: mul r1, 0x00000010
  0x28,0x05,0x00,0x00,0x00, // +05: mul r2, 0x00000005
  0x19, // +0A: add r2, r1
  0x44,0xcd,0xab,0x00,0x00, // +0B: xor r1, 0x0000ABCD
  0x16, // +10: add r1, r2
  0x5c,0xef,0xbe,0xad,0xde, // +11: mov r3, 0xDEADBEEF
  0x47, // +16: xor r1, r3
  0x36, // +17: sub r1, r2
  0x2b, // +18: mul r2, r3
  0x16, // +19: add r1, r2
};

//--------------------------------------------------------------------------
static const unsigned char x2[] =
{
  0x54,0x01,0x00,0x00,0x00, // +00: mov r1, 0x00000001
  0x58,0x05,0x00,0x00,0x00, // +05: mov r2, 0x00000005
  0x28,0x02,0x00,0x00,0x00, // +0A: mul r2, 0x00000002
  0x16, // +0F: add r1, r2
};

//--------------------------------------------------------------------------
static const char p_str1[] = {5, "Hello"};
static const char p_str2[] = {27, "This is a simple VM program"};

//--------------------------------------------------------------------------
bool vm_evaluate(
  const unsigned char *bytecode, 
  size_t sz, 
  unsigned long &r1, 
  unsigned long &r2, 
  unsigned long &r3)
{
  // Initialize the machine registers
  unsigned long regs[] = {0, r1, r2, r3};

  for (const unsigned char *p_end = bytecode + sz;bytecode < p_end;)
  {
    unsigned char op = *bytecode++;
    unsigned char inst = op >> 4;
    unsigned char op1 = (op & 0xf) >> 2;
    if (op1 == 0)
      return false;

    unsigned char op2 = op & 3;
    unsigned long imm;

    if (op2 == 0)
    {
      imm = *(unsigned long *)bytecode;
      bytecode += sizeof(unsigned long);
    }
    else
      imm = regs[op2];

    switch (inst)
    {
    case 1:
      // add
      regs[op1] += imm;
      break;
    case 2:
      // mul
      regs[op1] *= imm;
      break;
    case 3:
      // sub
      regs[op1] -= imm;
      break;
    case 4:
      // xor
      regs[op1] ^= imm;
      break;
    case 5:
      // mov
      regs[op1] = imm;
      break;
    }
  }
  r1 = regs[1];
  r2 = regs[2];
  r3 = regs[3];
  return true;
}

//--------------------------------------------------------------------------
bool vm_evaluate_obfs(
  unsigned long *r1, 
  unsigned long *r2, 
  unsigned long *r3,
  ...)
{
  va_list va;
  va_start(va, r3);
  std::vector<unsigned char> bytes;
  size_t sentinel = 0;
  while (true)
  {
    unsigned char b = (unsigned char)(va_arg(va, int) & 0xFF);
    if (b == 0xff)
    {
      sentinel++;
      if (sentinel == 5)
        break;
    }
    else
      sentinel = 0;
    bytes.push_back(b);
  }
  bool ok = vm_evaluate(&bytes[0], bytes.size()-4, *r1, *r2, *r3);
  va_end(va);
  return ok;
}

//--------------------------------------------------------------------------
void print_pascal_string(const char *str)
{
  int len = *(unsigned char *)str++ & 0xff;
  for (int i=0;i<len;i++)
    printf("%c", *str++);
  printf("\n");
}

//--------------------------------------------------------------------------
void print_rsrc_string(UINT id)
{
  char buf[1024];
  LoadStringA(g_hinst, id, buf, sizeof(buf));
  printf("%s\n", buf);
}

//--------------------------------------------------------------------------
unsigned long f_of_ab(unsigned long a, unsigned long b)
{
  const unsigned char *loc;
  unsigned long sz;
  __asm
  {
    push eax
    push ebx
    call L
    __emit 0x25 // +00: mul r1, r1
    __emit 0x2a // +01: mul r2, r2
    __emit 0x1d // +02: add r3, r1
    __emit 0x1e // +03: add r3, r2
    __emit 0x49 // +04: xor r2, r1
    __emit 0x1e // +05: add r3, r2  

    __emit 0x25  // +00: mul r1, r1
    __emit 0x2a  // +01: mul r2, r2
    __emit 0x1d  // +02: add r3, r1
    __emit 0x1e  // +03: add r3, r2
    __emit 0x49  // +04: xor r2, r1
    __emit 0x1e  // +05: add r3, r2
    __emit 0x1c  // +06: add r3, 0x45811900
    __emit 0x00
    __emit 0x19
    __emit 0x81
    __emit 0x45  
    __emit 0x57  // +0B: mov r1, r3
   L:
    call L1
   L1:
    pop ebx
    pop eax
    sub ebx, eax
    sub ebx, 5
    mov [sz], ebx
    mov [loc], eax
    pop ebx
    pop eax
  }
  unsigned long r3 = 0;
  vm_evaluate(loc, sz, a, b, r3);
  return a;
}

//--------------------------------------------------------------------------
// Some dummy code to exercise all the functions
int main(int argc, char *argv[])
{
  unsigned long r1=0, r2=0, r3=0;
  g_hinst = GetModuleHandle(argv[0]);

  print_pascal_string(p_str1);
  if (argc != 5)
  {
    r1 = 1;
    r2 = 3;
    r3 = 4;
    vm_evaluate(x1, sizeof(x1), r1, r2, r3);
    vm_evaluate_obfs(&r1, &r2, &r3, 
      0x4b, // +00: xor r2, r3
      0x17, // +01: add r1, r3
      0x16, // +02: add r1, r2
      0xff,0xff,0xff,0xff,0xff);
  }
  else
  {
    vm_evaluate(x2, sizeof(x2), r1, r2, r3);
  }

  printf("f_of_ab=%d\n", f_of_ab(5, 6 + argc));

  printf("r1=%lu r2=%lu r3=%lu\n", r1, r2, r3);
  print_pascal_string(p_str2);

  print_rsrc_string(IDS_STRING101);
  print_rsrc_string(IDS_STRING102);
  print_rsrc_string(IDS_STRING103);
  return 0;
}
